from abc import ABC, abstractmethod
from threading import Lock

from eric_sse.message import Message
from eric_sse.exception import NoMessagesException

class RepositoryError(Exception):
    ...

class Queue(ABC):
    @abstractmethod
    def pop(self) -> Message:
        ...

    @abstractmethod
    def push(self, message: Message) -> None:
        ...

    @abstractmethod
    def delete(self) -> None:
        ...

class InMemoryQueue(Queue):
    def __init__(self):
        self.__messages: list[Message] = []

    def pop(self) -> Message:
        try:
            with Lock():
                return self.__messages.pop(0)
        except IndexError:
            raise NoMessagesException

    def push(self, message: Message) -> None:
        self.__messages.append(message)

    def delete(self) -> None:
        self.__messages = []

class AbstractMessageQueueFactory(ABC):
    """
    Abstraction for queues creation

    see :class:`eric_sse.entities.AbstractChannel`
    """
    @abstractmethod
    def create(self) -> Queue:
        ...


class InMemoryMessageQueueFactory(AbstractMessageQueueFactory):
    """
    Default implementation used by :class:`eric_sse.entities.AbstractChannel`
    """
    def create(self) -> Queue:
        return InMemoryQueue()



from dataclasses import dataclass


@dataclass
class Message:
    """
    Models a message

    It's just a container of information identified by a type.
    For validation purposes you can override MessageQueueListener.on_message
    """
    type: str
    payload: dict | list | str | int | float | None = None


@dataclass
class UniqueMessage:
    __message: Message
    __sender_id: str | None = None

    def __init__(self, message_id: str, message: Message, sender_id: str = None) -> None:
        self.__id = message_id
        self.__message = message
        self.__sender_id = sender_id

    @property
    def id(self) -> str:
        return str(self.__id)

    @property
    def type(self) -> str:
        return  self.__message.type

    @property
    def sender_id(self) -> str:
        return self.__sender_id

    @property
    def payload(self) -> dict | list | str | int | float | None:
        return self.__message.payload


@dataclass
class SignedMessage(Message):
    """A wrapper that adds sender id"""

    def __init__(self, sender_id: str, msg_type: str, msg_payload: dict | list | str | int | float | None = None):
        self.__sender_id = sender_id
        self.__msg_type = msg_type
        self.__msg_payload = msg_payload

    @property
    def sender_id(self) -> str:
        return self.__sender_id

    @property
    def type(self):
        return self.__msg_type

    @property
    def payload(self) -> dict | list | str | int | float | None:
        return {'sender_id': self.__sender_id, 'payload': self.__msg_payload}
